"""
mediapipe
- author: Auorui (夏天是冰红茶)
- creation time:2022.5
- pyzjr.mediapipe is a part of the model strategy used for deep learning
- WeChat: z15583909992
- Email: zjricetea@gmail.com
- Note: Currently still being updated, please refer to the latest version for any changes that may occur
"""


from .HandTrack import Hands


__all__ = ["Hands",]