from .Standardpooling import MaxPool2d, AvgPool2d
from .HaarWavelet import HWDownsampling
from .StrideConv import StridedConvolution
from .strippooling import StripPooling



__all__ = ["MaxPool2d", "AvgPool2d", "HWDownsampling", "StridedConvolution", "StripPooling"]