"""
Copyright (c) 2023, Auorui.
All rights reserved.

大多是考研数学所用的,闲来无事写的
"""
from .point_line import *
from .compute import *
from .arithmetic import *
from .constant import *
from .function import *
from .Numpy import *
