"""
- author: Auorui (夏天是冰红茶)
- creation time:2022.10
- pyzjr is the Codebase accumulated by my python programming.
  At present, it is only for my personal use. If you want to
  use it, please contact me. Here are my email and WeChat.
- WeChat: z15583909992
- Email: zjricetea@gmail.com / 2165648225@qq.com
- Note: Currently still being updated, please refer to the latest version for any changes that may occur

    https://www.pepy.tech/projects/pyzjr
    API documentation website:
"""
from pyzjr.utils import *

from pyzjr.Z import * 
from pyzjr.Math import *
from pyzjr.PIC import *
from pyzjr.data import *
from pyzjr.devices import *

from pyzjr.augmentation import *
from pyzjr.mediapipe import *
from pyzjr.dlearn import *
from pyzjr.core import *
from pyzjr.videos import *
from pyzjr.version import __version__

from pyzjr.nn import *
from pyzjr.Models import *