"""
Copyright (c) 2023, Auorui.
All rights reserved.

The following citation imports roughly follow the sorting by year
"""
# 1990s
from .lenet import LeNet

# 2010s
from .alexnet import AlexNet
from .vgg import VGG, vgg11_bn, vgg13_bn, vgg16_bn, vgg19_bn
from .googlenet import GoogLeNet
from .resnet import ResNet, resnet18, resnet34, resnet50, resnet101, resnet152
from .squeezenet import SqueezeNet, squeezenet1_0, squeezenet1_1
from .se_resnet import se_resnet18, se_resnet34, se_resnet50, se_resnet101, se_resnet152
from .darknet import darknet19, darknet53
from .mobilenet import MobileNetV1, MobileNetV2, MobileNetV3, MobileNetV3_Large, MobileNetV3_Small
from .densenet import DenseNet, densenet121, densenet161, densenet169, densenet201
from .mnasnet import MNASNet, mnasnet0_5, mnasnet0_75, mnasnet1_0, mnasnet1_3

# 2020s
from .conv2former import Conv2Former_n, Conv2Former_t, Conv2Former_s, Conv2Former_l, Conv2Former_b
from .ghostnet import GhostNet, GGhostRegNet, ghostnetv1, ghostnetv2, g_ghost_regnetx_002, \
    g_ghost_regnetx_004, g_ghost_regnetx_006, g_ghost_regnetx_008, g_ghost_regnetx_016, \
    g_ghost_regnetx_032, g_ghost_regnetx_040, g_ghost_regnetx_064, g_ghost_regnetx_080, \
    g_ghost_regnetx_120, g_ghost_regnetx_160, g_ghost_regnetx_320
from .regnet import regnetx_002, regnetx_004, regnetx_006, regnetx_008, regnetx_016, regnetx_032,\
    regnetx_040, regnetx_064, regnetx_080, regnetx_120, regnetx_160, regnetx_320
from .fasternet import fasternet_t0, fasternet_t1, fasternet_t2, fasternet_s, \
    fasternet_m, fasternet_l
